package cz.muni.fi.pa165.service.band;

import cz.muni.fi.pa165.data.entity.Album;
import cz.muni.fi.pa165.data.entity.Band;
import cz.muni.fi.pa165.data.entity.Tour;
import cz.muni.fi.pa165.data.repository.album.AlbumRepository;
import cz.muni.fi.pa165.data.repository.band.BandRepository;
import cz.muni.fi.pa165.data.repository.tour.TourRepository;
import cz.muni.fi.pa165.event.BandDeletedEvent;
import cz.muni.fi.pa165.event.TeamMemberCreatedEvent;
import cz.muni.fi.pa165.event.TeamMemberDeletedEvent;
import cz.muni.fi.pa165.exception.model.AlbumNotFoundException;
import cz.muni.fi.pa165.exception.model.BandNotFoundException;
import cz.muni.fi.pa165.exception.model.TourNotFoundException;
import cz.muni.fi.pa165.mapper.BandMapper;
import cz.muni.fi.pa165.model.band.BandDTO;
import cz.muni.fi.pa165.model.band.BandCreateDTO;
import jakarta.persistence.EntityNotFoundException;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.*;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.security.core.context.SecurityContextHolder;
import org.springframework.security.oauth2.server.resource.authentication.BearerTokenAuthentication;
import org.springframework.stereotype.Service;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestTemplate;

import java.util.ArrayList;
import java.util.List;

@Service
@Slf4j
public class BandServiceImpl implements BandService {

    private final BandRepository bandRepository;

    private final AlbumRepository albumRepository;

    private final TourRepository tourRepository;

    private final BandMapper bandMapper;

    private final KafkaTemplate<String, BandDeletedEvent> kafkaTemplate;

    private final RestTemplate restTemplate;

    @Autowired
    public BandServiceImpl(BandRepository bandRepository, AlbumRepository albumRepository, TourRepository tourRepository, BandMapper bandMapper, KafkaTemplate<String, BandDeletedEvent> kafkaTemplate, RestTemplate restTemplate) {
        this.bandRepository = bandRepository;
        this.albumRepository = albumRepository;
        this.tourRepository = tourRepository;
        this.bandMapper = bandMapper;
        this.kafkaTemplate = kafkaTemplate;
        this.restTemplate = restTemplate;
    }

    @Override
    public BandDTO createBand(BandCreateDTO bandCreateDTO) {
        long bandManagerId = bandCreateDTO.getBandManagerId();

        BearerTokenAuthentication authentication = (BearerTokenAuthentication) SecurityContextHolder.getContext().getAuthentication();
        String accessToken = authentication.getToken().getTokenValue();
        HttpHeaders headers = new HttpHeaders();
        headers.setBearerAuth(accessToken);
        HttpEntity<String> entity = new HttpEntity<>(headers);

        // Check if the band manager exists
        try {
            ResponseEntity<Object> bandManagerResponseEntity = restTemplate.exchange("http://pa165-band-manager-service:9094/api/v1/band-manager/{bandManagerId}", HttpMethod.GET, entity, Object.class, bandManagerId);
        } catch (HttpClientErrorException e) {
            if (e.getStatusCode().equals(HttpStatus.NOT_FOUND)) {
                throw new EntityNotFoundException("Band manager with ID " + bandManagerId + " was not found");            }
        }

        Band created = bandRepository.save(
                bandMapper.bandCreateDTOToBand(bandCreateDTO)
        );

        return bandMapper.bandToBandDTO(created);
    }

    @Override
    public BandDTO getBandById(long id) {
        Band found = bandRepository.findById(id).orElseThrow(
                () -> new BandNotFoundException(String.format("Band with id %s was not found", id))
        );
        return bandMapper.bandToBandDTO(found);
    }

    @Override
    public BandDTO updateBandById(long id, BandDTO bandDTO) {
        bandDTO.setId(id);

        Band mappedBand = bandMapper.bandDTOToBand(bandDTO);

        // Find corresponding albums and add them
        List<Album> foundAlbums = new ArrayList<>();
        bandDTO.getAlbumIds().forEach(
                albumId -> foundAlbums.add(albumRepository.findById(albumId).orElseThrow(
                        () -> new AlbumNotFoundException(String.format("Album with id %s was not found", albumId))
                ))
        );
        mappedBand.setAlbums(foundAlbums);

        // Find corresponding tours and add them
        List<Tour> foundTours = new ArrayList<>();
        bandDTO.getTourIds().forEach(
                tourId -> foundTours.add(tourRepository.findById(tourId).orElseThrow(
                        () -> new TourNotFoundException(String.format("Tour with id %s was not found", tourId))
                ))
        );
        mappedBand.setTours(foundTours);

        // Update the band
        Band saved = bandRepository.save(mappedBand);

        return bandMapper.bandToBandDTO(saved);
    }

    @Override
    public void removeBandById(long id) {
        bandRepository.deleteById(id);
        var event = new BandDeletedEvent(id);
        kafkaTemplate.send("bandDeletedTopic", event);
        log.info("Sending band deleted event: {}", event);
    }

    @Override
    public void addTeamMemberToBand(long id, long teamMemberId) {
        BandDTO bandDTO = getBandById(id);
        bandDTO.getBandMemberIds().add(teamMemberId);
        updateBandById(id, bandDTO);
    }

    public void removeTeamMemberFromBand(long id, long teamMemberId) {
        BandDTO bandDTO = getBandById(id);
        bandDTO.getBandMemberIds().remove(teamMemberId);
        updateBandById(id, bandDTO);
    }

    @KafkaListener(topics = "teamMemberCreatedTopic", autoStartup = "true", groupId = "teamMemberCreatedId")
    public void handleProposalAcceptedEvent(TeamMemberCreatedEvent teamMemberCreatedEvent) {
        log.info("Received team member created event: {}", teamMemberCreatedEvent);
        addTeamMemberToBand(teamMemberCreatedEvent.getBandId(), teamMemberCreatedEvent.getTeamMemberId());
    }

    @KafkaListener(topics = "teamMemberDeletedTopic", autoStartup = "true", groupId = "teamMemberDeletedId")
    public void handleUserDeletedEvent(TeamMemberDeletedEvent teamMemberDeletedEvent) {
        log.info("Received team member deleted event: {}", teamMemberDeletedEvent);
        removeTeamMemberFromBand(teamMemberDeletedEvent.getBandId(), teamMemberDeletedEvent.getTeamMemberId());
    }

}
