package cz.muni.fi.pa165.service.album;

import cz.muni.fi.pa165.data.entity.Album;
import cz.muni.fi.pa165.data.entity.Band;
import cz.muni.fi.pa165.data.entity.Song;
import cz.muni.fi.pa165.data.repository.album.AlbumRepository;
import cz.muni.fi.pa165.data.repository.band.BandRepository;
import cz.muni.fi.pa165.data.repository.song.SongRepository;
import cz.muni.fi.pa165.exception.model.AlbumNotFoundException;
import cz.muni.fi.pa165.exception.model.BandNotFoundException;
import cz.muni.fi.pa165.exception.model.SongNotFoundException;
import cz.muni.fi.pa165.mapper.AlbumMapper;
import cz.muni.fi.pa165.model.album.AlbumCreateDTO;
import cz.muni.fi.pa165.model.album.AlbumDTO;
import org.junit.jupiter.api.Assertions;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyLong;
import static org.mockito.Mockito.*;

public class AlbumServiceImplTest {

    @Mock
    private AlbumRepository albumRepository;

    @Mock
    private SongRepository songRepository;

    @Mock
    private BandRepository bandRepository;

    @Mock
    private AlbumMapper albumMapper;

    private AlbumServiceImpl albumService;

    @BeforeEach
    void setUp() {
        MockitoAnnotations.openMocks(this);
        albumService = new AlbumServiceImpl(albumRepository, songRepository, bandRepository, albumMapper);
    }

    @Test
    void testGetAlbumById() {
        long albumId = 1L;
        Album album = new Album();
        album.setId(albumId);
        when(albumRepository.findById(albumId)).thenReturn(Optional.of(album));
        AlbumDTO expectedAlbumDTO = new AlbumDTO();
        expectedAlbumDTO.setId(albumId);
        when(albumMapper.albumToAlbumDTO(album)).thenReturn(expectedAlbumDTO);

        AlbumDTO result = albumService.getAlbumById(albumId);

        assertEquals(expectedAlbumDTO, result);
        Mockito.verify(albumRepository).findById(albumId);
        Mockito.verify(albumMapper).albumToAlbumDTO(album);
    }

    @Test
    void testUpdateAlbumById() {
        long albumId = 1L;
        long songId1 = 1L;
        long songId2 = 2L;

        AlbumDTO albumDTO = new AlbumDTO();
        albumDTO.setId(albumId);
        albumDTO.setName("Test Album");
        albumDTO.setBandId(1L);
        albumDTO.setSongIds(Arrays.asList(songId1, songId2));

        Song song1 = new Song();
        song1.setId(songId1);

        Song song2 = new Song();
        song2.setId(songId2);

        List<Song> songs = Arrays.asList(song1, song2);

        Album album = new Album();
        album.setId(albumId);
        album.setName("Test Album");
        album.setBand(new Band());
        album.setSongs(songs);

        Album mappedAlbum = new Album();
        mappedAlbum.setId(albumId);
        mappedAlbum.setName("Test Album");
        mappedAlbum.setBand(new Band());
        mappedAlbum.setSongs(songs);

        when(albumMapper.albumDTOToAlbum(albumDTO)).thenReturn(mappedAlbum);
        when(albumMapper.albumToAlbumDTO(album)).thenReturn(albumDTO);
        when(songRepository.findById(songId1)).thenReturn(Optional.of(song1));
        when(songRepository.findById(songId2)).thenReturn(Optional.of(song2));
        when(albumRepository.save(mappedAlbum)).thenReturn(album);

        AlbumDTO result = albumService.updateAlbumById(albumId, albumDTO);

        assertEquals(albumId, result.getId());
        assertEquals("Test Album", result.getName());
        assertEquals(2, result.getSongIds().size());
        assertTrue(result.getSongIds().contains(songId1));
        assertTrue(result.getSongIds().contains(songId2));

        Mockito.verify(albumMapper).albumDTOToAlbum(albumDTO);
        Mockito.verify(songRepository).findById(songId1);
        Mockito.verify(songRepository).findById(songId2);
        Mockito.verify(albumRepository).save(mappedAlbum);
    }

    @Test
    void updateAlbumByIdSongNotFound() {
        long id = 1;
        AlbumDTO albumDTO = new AlbumDTO();
        albumDTO.setId(id);
        albumDTO.setSongIds(Arrays.asList(1L, 2L));
        when(songRepository.findById(1L)).thenReturn(Optional.of(new Song()));
        when(songRepository.findById(2L)).thenReturn(Optional.empty());

        assertThrows(SongNotFoundException.class, () -> albumService.updateAlbumById(id, albumDTO));

        verify(songRepository, times(1)).findById(1L);
        verify(songRepository, times(1)).findById(2L);
    }

    @Test
    void testRemoveAlbumById() {
        long albumId = 1L;
        Album album = new Album();
        album.setId(albumId);
        when(albumRepository.findById(albumId)).thenReturn(Optional.of(album));

        albumService.removeAlbumById(albumId);

        Mockito.verify(albumRepository).findById(albumId);
        Mockito.verify(albumRepository).delete(album);
    }
}