package cz.muni.fi.pa165.service.tour;

import cz.muni.fi.pa165.data.entity.Band;
import cz.muni.fi.pa165.data.entity.Tour;
import cz.muni.fi.pa165.data.entity.TourStop;
import cz.muni.fi.pa165.data.repository.band.BandRepository;
import cz.muni.fi.pa165.data.repository.tour.TourRepository;
import cz.muni.fi.pa165.data.repository.tourstop.TourStopRepository;
import cz.muni.fi.pa165.exception.model.TourNotFoundException;
import cz.muni.fi.pa165.exception.model.TourStopNotFoundException;
import cz.muni.fi.pa165.mapper.TourMapper;
import cz.muni.fi.pa165.mapper.TourStopMapper;
import cz.muni.fi.pa165.model.tour.TourCreateDTO;
import cz.muni.fi.pa165.model.tour.TourDTO;
import cz.muni.fi.pa165.model.tour.TourStopDTO;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.BeforeEach;
import org.mockito.Mock;
import org.mockito.Mockito;
import org.mockito.MockitoAnnotations;

import java.util.Optional;

import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

public class TourServiceImplTest {

    @Mock
    private TourRepository tourRepository;

    @Mock
    private TourStopRepository tourStopRepository;

    @Mock
    private BandRepository bandRepository;

    @Mock
    private TourMapper tourMapper;

    @Mock
    private TourStopMapper tourStopMapper;

    private TourServiceImpl tourService;

    @BeforeEach
    void setUp() {
        MockitoAnnotations.openMocks(this);
        tourService = new TourServiceImpl(tourRepository, tourStopRepository, bandRepository, tourMapper, tourStopMapper);
    }

    @Test
    public void testCreateTour() {
        TourCreateDTO tourCreateDTO = new TourCreateDTO();
        tourCreateDTO.setBandId(1L);
        Tour tour = new Tour();
        tour.setId(1L);
        tour.setBand(new Band());
        TourDTO expected = new TourDTO();
        expected.setId(1L);
        when(tourMapper.tourCreateDTOToTour(tourCreateDTO)).thenReturn(tour);
        when(bandRepository.findById(1L)).thenReturn(Optional.of(new Band()));
        when(tourRepository.save(tour)).thenReturn(tour);
        when(tourMapper.tourToTourDTO(tour)).thenReturn(expected);

        TourDTO result = tourService.createTour(tourCreateDTO);

        verify(tourMapper).tourCreateDTOToTour(tourCreateDTO);
        verify(bandRepository).findById(1L);
        verify(tourRepository).save(tour);
        verify(tourMapper).tourToTourDTO(tour);
        assertEquals(expected, result);
    }

    @Test
    public void testGetTourById() {
        // Given
        long id = 1L;
        Tour tour = new Tour();
        tour.setId(id);
        TourDTO expected = new TourDTO();
        expected.setId(id);
        when(tourRepository.findById(id)).thenReturn(Optional.of(tour));
        when(tourMapper.tourToTourDTO(tour)).thenReturn(expected);

        // When
        TourDTO result = tourService.getTourById(id);

        // Then
        verify(tourRepository).findById(id);
        verify(tourMapper).tourToTourDTO(tour);
        assertEquals(expected, result);
    }

    @Test
    public void testUpdateTourById() {
        long id = 1L;
        TourDTO tourDTO = new TourDTO();
        tourDTO.setId(id);
        Tour tour = new Tour();
        tour.setId(id);

        when(tourRepository.save(Mockito.any(Tour.class))).thenReturn(tour);
        when(tourMapper.tourToTourDTO(tour)).thenReturn(tourDTO);
        when(tourMapper.tourDTOToTour(tourDTO)).thenReturn(tour);

        TourDTO result = tourService.updateTourById(id, tourDTO);

        verify(tourRepository).save(tour);
        verify(tourMapper).tourToTourDTO(tour);
        assertEquals(tourDTO, result);
    }

    @Test
    public void testRemoveTourById() {
        long id = 1L;
        Tour tour = new Tour();
        tour.setId(id);
        when(tourRepository.findById(id)).thenReturn(Optional.of(tour));

        tourService.removeTourById(id);

        verify(tourRepository).findById(id);
        verify(tourRepository).delete(tour);
    }

    @Test
    public void testCreateTourStop() {
        // Given
        long tourId = 1L;
        long tourStopId = 2L;
        Tour tour = new Tour();
        tour.setId(tourId);
        TourStopDTO tourStopDTO = new TourStopDTO();
        tourStopDTO.setId(tourStopId);
        TourStop tourStop = new TourStop();
        tourStop.setId(tourStopId);
        when(tourRepository.findById(tourId)).thenReturn(Optional.of(tour));
        when(tourStopMapper.tourStopDTOToTourStop(tourStopDTO)).thenReturn(tourStop);
        when(tourStopRepository.save(tourStop)).thenReturn(tourStop);
        when(tourStopMapper.tourStopToTourStopDTO(tourStop)).thenReturn(tourStopDTO);

        // When
        TourStopDTO createdTourStop = tourService.createTourStop(tourId, tourStopDTO);

        // Then
        verify(tourRepository, times(1)).findById(tourId);
        verify(tourStopMapper, times(1)).tourStopDTOToTourStop(tourStopDTO);
        verify(tourStopRepository, times(1)).save(tourStop);
        verify(tourStopMapper, times(1)).tourStopToTourStopDTO(tourStop);
    }

    @Test
    public void testCreateTourStopTourNotFound() {
        // Given
        long tourId = 1L;
        TourStopDTO tourStopDTO = new TourStopDTO();
        when(tourRepository.findById(tourId)).thenReturn(Optional.empty());

        assertThrows(TourNotFoundException.class, () -> tourService.createTourStop(tourId, tourStopDTO));
    }

    @Test
    public void testDeleteTourStop() {
        // Given
        long tourId = 1L;
        long tourStopId = 2L;
        Tour tour = new Tour();
        tour.setId(tourId);
        when(tourRepository.findById(tourId)).thenReturn(Optional.of(tour));
        TourStop tourStop = new TourStop();
        when(tourStopRepository.findById(tourStopId)).thenReturn(Optional.of(tourStop));

        TourDTO deletedTour = tourService.deleteTourStop(tourId, tourStopId);

        verify(tourRepository, times(1)).findById(tourId);
        verify(tourStopRepository, times(1)).findById(tourStopId);
        verify(tourStopRepository, times(1)).delete(tourStop);
        verify(tourMapper, times(1)).tourToTourDTO(tour);
    }

    @Test
    void testDeleteTourStopTourStopNotFound() {
        long tourId = 1L;
        long tourStopId = 1L;
        when(tourStopRepository.findById(tourStopId)).thenReturn(Optional.empty());

        assertThrows(TourStopNotFoundException.class, () -> tourService.deleteTourStop(tourId, tourStopId));

        verify(tourStopRepository).findById(tourStopId);
        verifyNoMoreInteractions(tourStopRepository);
    }
}